/**The MIT License (MIT)

Copyright (c) 2015 by Daniel Eichhorn

Changed/updated by:
Main changes are
1) Driver for OLED is for SDD1306 chip
2) Some more data fields added to the Weather
3) Commented out "Thingspeak" function
4) Added a shutown timer (and a reset/wake button)
5) Added a shaker vibration switch to toggle weather location

Copyright (c) 2017 by Simon Jowett

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

See more at http://blog.squix.ch
*/


#include <Ticker.h>
#include <JsonListener.h>
#include "SDD1306.h"
#include "SDD1306Ui.h"
#include "Wire.h"
#include <ESP8266WiFi.h>
#include "WundergroundClient.h"
#include "WeatherStationFonts.h"
#include "WeatherStationImages.h"
#include "TimeClient.h"
//#include "ThingspeakClient.h"
#include "DHT.h"




/***************************
 * Begin Settings
 **************************/

// WIFI
const char* WIFI_SSID = "Your SSID";
const char* WIFI_PWD = "Your Passphrase";


// Setup
const int UPDATE_INTERVAL_SECS = 10 * 60; // Update every 10 minutes

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 12;
const int SDC_PIN = 13;

// TimeClient settings
const float UTC_OFFSET = 0;
float utcoh = 0;

// Wunderground Settings
const boolean IS_METRIC = true;
const String WUNDERGRROUND_API_KEY = "Your WU api key";
String WUNDERGROUND_COUNTRY = "UK";
String WUNDERGROUND_CITY = "Leeds_Bradford";

// NB!!! - if the city has a 'space' in it ie Cape Town, use an underscore _
const String city1 = "Leeds_Bradford";
const String country1 = "UK";

const String city2 = "Sydney";
const String country2 = "AU";

//Thingspeak Settings
//const String THINGSPEAK_CHANNEL_ID = "67284";
//const String THINGSPEAK_API_READ_KEY = "L2VIW20QVNZJBLAK";

// Initialize the oled display for address 0x3c or 0x3d
// sda-pin and sdc-pin
SDD1306   display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);
SDD1306Ui ui     ( &display );

// DHT Settings
#define DHTPIN 14     // what digital pin we're connected to.

// Uncomment whatever type you're using!
//#define DHTTYPE DHT11   // DHT 11
#define DHTTYPE DHT22   // DHT 22  (AM2302), AM2321
//#define DHTTYPE DHT21   // DHT 21 (AM2301)

#define shakerPin 5         // what digital pin the shaker switch is connect to
volatile int state = LOW;   // Volatile to toggls low to high and back when vibration sensor is shaken
int currState = state;      // variable to compare to see if the state has changed
volatile long lastVibration = millis(); // required to debounce the shaker.

unsigned long timerSleep = 0; // timer for esp8266 shutdown

/***************************
 * End Settings
 **************************/

TimeClient timeClient(UTC_OFFSET);

// Set to false, if you prefere imperial/inches, Fahrenheit
WundergroundClient wunderground(IS_METRIC);

//ThingspeakClient thingspeak;

// Initialize the temperature/ humidity sensor
DHT dht(DHTPIN, DHTTYPE);

// this array keeps function pointers to all frames
// frames are the single views that slide from right to left
bool (*frames[])(SDD1306 *display, SDD1306UiState* state, int x, int y)  { drawFrame1, drawFrame2, drawFrame3, drawFrame4, drawFrame6 };
int numberOfFrames = 5;

// flag changed in the ticker function every 10 minutes
bool readyForWeatherUpdate = true;

String lastUpdate = "--";

float humidity = 0.;
float temperature = 0.;

Ticker ticker;

void setup() {

  Serial.begin(115200);
  Serial.println();
  Serial.println();

  // initialize dispaly
  display.init();
  display.clear();
  display.display();

  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.setContrast(255);

  wificonnect();

  ui.setTargetFPS(30);

  ui.setActiveSymbole(activeSymbole);
  ui.setInactiveSymbole(inactiveSymbole);

  // You can change this to
  // TOP, LEFT, BOTTOM, RIGHT
  ui.setIndicatorPosition(BOTTOM);

  // Defines where the first frame is located in the bar.
  ui.setIndicatorDirection(LEFT_RIGHT);

  // You can change the transition that is used
  // SLIDE_LEFT, SLIDE_RIGHT, SLIDE_TOP, SLIDE_DOWN
  ui.setFrameAnimation(SLIDE_LEFT);

  // Add frames
  ui.setFrames(frames, numberOfFrames);

  // Inital UI takes care of initalising the display too.
  ui.init();
  display.flipScreenVertically();

  Serial.println("");

 // updateData(&display);

  ticker.attach(UPDATE_INTERVAL_SECS, setReadyForWeatherUpdate);
  
  pinMode(shakerPin, INPUT_PULLUP);             // Vibration sensor input 
  attachInterrupt(shakerPin, shaker, FALLING);  // Set an inturrupt that will toggle the city location when the vibration switch is shaken
}

void loop() {

  if (readyForWeatherUpdate && ui.getUiState().frameState == FIXED) {
    if (WiFi.status() != WL_CONNECTED) {
      wificonnect();
    }
    if (WiFi.status() == WL_CONNECTED){
      updateData(&display);
    }
  }

  int remainingTimeBudget = ui.update();

  if (remainingTimeBudget > 0) {
    // You can do some work here
    // Don't do stuff if you are below your
    // time budget.
    //read the pushbutton value into a variable

    delay(remainingTimeBudget);
  }

    if (state != currState) {
      Serial.println(state);
      currState = state;
      changecity();
  }

   // Shutdown and go to sleep function
  
  timerSleep = millis();
  if (timerSleep >= 10*60000 && ui.getUiState().frameState == FIXED){ // after 2 minutes go to sleep
    drawSleep(&display);
        // go to deepsleep for xx minutes or 0 = permanently
        ESP.deepSleep(0,  WAKE_RF_DEFAULT);                       // 0 delay = permanently to sleep
        delay(1000);                                              // delay to allow the ESP to go to sleep.
  }
}

void updateData(SDD1306 *display) {


  drawProgress(display, 10, "Updating conditions...");
  wunderground.updateConditions(WUNDERGRROUND_API_KEY, WUNDERGROUND_COUNTRY, WUNDERGROUND_CITY);

  drawProgress(display, 30, "Updating forecasts...");
  wunderground.updateForecast(WUNDERGRROUND_API_KEY, WUNDERGROUND_COUNTRY, WUNDERGROUND_CITY);

  //drawProgress(display, 50, "Updating thingspeak...");
  //thingspeak.getLastChannelItem(THINGSPEAK_CHANNEL_ID, THINGSPEAK_API_READ_KEY);
  //lastUpdate = timeClient.getFormattedTime();

  setUTCOffset();

  drawProgress(display, 80, "Updating time...");
  timeClient.updateTime();

  drawProgress(display, 90, "Updating local temperature and humidity");
  humidity = dht.readHumidity();
  // Read temperature as Fahrenheit (isFahrenheit = true)
  temperature = dht.readTemperature(!IS_METRIC);
  drawProgress(display, 100, "Done...");
  readyForWeatherUpdate = false;
  delay(1000);


}

void drawProgress(SDD1306 *display, int percentage, String label) {
  display->clear();
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64, 0, "Getting data for " + WUNDERGROUND_CITY);
  display->drawString(64, 10, label);
  display->drawRect(10, 28, 108, 12);
  display->fillRect(12, 30, 104 * percentage / 100 , 9);
  display->display();
}


bool drawFrame1(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64 + x, 0 + y, "Time in " + WUNDERGROUND_CITY);
  String date = wunderground.getDate();
  int textWidth = display->getStringWidth(date);
  display->drawString(64 + x, 10 + y, date);
  display->setFont(ArialMT_Plain_24);
  String time = timeClient.getFormattedTime();
  int timehour = time.substring(0, 2).toInt();
  // Serial.println(timehour);
  timehour = timehour + utcoh;
  if (timehour >= 24 ) { // remove 24 hrs if hours is > 24
    timehour = timehour - 24;
  }
  // Serial.println(timehour);
  String timeh = String(timehour);
  if (timehour < 10 ) { // add a leading zero if the hours is less thsn 10
    timeh = "0" + timeh;
  }

  // Serial.println(timeh);
  textWidth = display->getStringWidth(time);
  display->drawString(64 + x, 20 + y, timeh + ":" + time.substring(3));
  display->setTextAlignment(TEXT_ALIGN_LEFT);

}

bool drawFrame2(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  display->setFont(ArialMT_Plain_10);
  display->drawString(49 + x, 10 + y, wunderground.getWeatherText());

  display->setFont(ArialMT_Plain_24);
  String temp = wunderground.getCurrentTemp() + "°C";
  display->drawString(49 + x, 20 + y, temp);
  int tempWidth = display->getStringWidth(temp);

  display->setFont(Meteocons_0_42);
  String weatherIcon = wunderground.getTodayIcon();
  int weatherIconWidth = display->getStringWidth(weatherIcon);
  display->drawString(25 + x - weatherIconWidth / 2, 10 + y, weatherIcon);
}

bool drawFrame3(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(32 + x, 0 + y, "Humidity");
  display->drawString(96 + x, 0 + y, "Pressure");
  display->drawString(32 + x, 28 + y, "Precipit.");
  display->drawString(96 + x, 28 + y, "UV index");

  display->setFont(ArialMT_Plain_16);
  display->drawString(32 + x, 10 + y, wunderground.getHumidity());
  display->drawString(96 + x, 10 + y, wunderground.getPressure());
  display->drawString(32 + x, 38 + y, wunderground.getPrecipitationToday());
  display->drawString(96 + x, 38 + y, wunderground.getUv());
}

bool drawFrame4(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  drawForecast(display, x, y, 0);
  drawForecast(display, x + 44, y, 2);
  drawForecast(display, x + 88, y, 4);
}
/*
bool drawFrame5(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64 + x, 0 + y, "Outdoor");
  display->setFont(ArialMT_Plain_24);
  display->drawString(64 + x, 10 + y, thingspeak.getFieldValue(0) + "°C");
  display->drawString(64 + x, 30 + y, thingspeak.getFieldValue(1) + "%");
}
*/
bool drawFrame6(SDD1306 *display, SDD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64 + x, 0 + y, "Indoor");
  display->setFont(ArialMT_Plain_24);
  display->drawString(64 + x, 10 + y, String(temperature) + "°C");
  display->drawString(64 + x, 30 + y, String(humidity) + "%");
}

void drawForecast(SDD1306 *display, int x, int y, int dayIndex) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  String day = wunderground.getForecastTitle(dayIndex).substring(0, 3);
  day.toUpperCase();
  display->drawString(x + 20, y, day);

  display->setFont(Meteocons_0_21);
  display->drawString(x + 20, y + 15, wunderground.getForecastIcon(dayIndex));

  display->setFont(ArialMT_Plain_16);
  display->drawString(x + 20, y + 37, wunderground.getForecastLowTemp(dayIndex) + "/" + wunderground.getForecastHighTemp(dayIndex));
  //display.drawString(x + 20, y + 51, );
  display->setTextAlignment(TEXT_ALIGN_LEFT);
}
void drawSleep(SDD1306 *display) {

        display->clear();
        display->setTextAlignment(TEXT_ALIGN_CENTER);
        display->setFont(ArialMT_Plain_10);
        display->drawString(64, 0, "Going to Sleep!");
        display->display();
        delay(5000);
        display->displayOff();
        Serial.println();
        Serial.println("closing connection. going to sleep...");
        delay(1000);
        timerSleep = 0;


}

void setReadyForWeatherUpdate() {
  Serial.println("Setting readyForUpdate to true");
  readyForWeatherUpdate = true;
}

void setUTCOffset() {
  String currUTCOffset = wunderground.getCurrUTCOffset();
  String currUTCOffsetSign = wunderground.getCurrUTCOffsetSign();
  Serial.print ("UTC Offset = ");
  Serial.println (currUTCOffset);
  Serial.print ("UTC Offset Sign = ");
  Serial.println (currUTCOffsetSign);
  utcoh = currUTCOffset.substring(0, 2).toInt();
  float utcom = currUTCOffset.substring(2, 4).toInt();
  if (currUTCOffsetSign == "-") {
    utcoh = utcoh * -1;
  }
  Serial.println (utcoh);  // number of hours offest from UTC
  Serial.println (utcom);   // number of minutes offset from utc (to capture the 30 minute time zones)
}

void shaker(){
  long timeNow = millis();

  if (timeNow - lastVibration > 10000){
  state=!state;
}
  lastVibration = timeNow;
}

void changecity(){
  // toggle city

  if (WiFi.status() != WL_CONNECTED)  {
    Serial.println("unable to toggle city data as wifi not connected");
    display.clear();
    display.setFont(ArialMT_Plain_10);
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(60, 10, "Unable to toggle city as");
    display.drawString(60, 20, "WiFi not connected");
    display.display();
    delay (6000);
    wificonnect();
  }
  else{
    if ( WUNDERGROUND_CITY == city1){
      WUNDERGROUND_CITY = city2;
      WUNDERGROUND_COUNTRY = country2;
    }
    else{
      WUNDERGROUND_CITY = city1;
      WUNDERGROUND_COUNTRY = country1;
    }
   readyForWeatherUpdate = true;
   Serial.println("Toggled city to " + WUNDERGROUND_CITY);
  }
}

void wificonnect(){

  WiFi.begin(WIFI_SSID, WIFI_PWD);
  long timer = millis(); // only search for a wifi for 30 seconds - stop burn in on OLED
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED && timer+10000 > millis()) {
    delay(500);
    Serial.print(".");
    display.clear();
    display.setFont(ArialMT_Plain_10);
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 10, "Connecting to WiFi");
    display.drawXbm(46, 30, 8, 8, counter % 3 == 0 ? activeSymbole : inactiveSymbole);
    display.drawXbm(60, 30, 8, 8, counter % 3 == 1 ? activeSymbole : inactiveSymbole);
    display.drawXbm(74, 30, 8, 8, counter % 3 == 2 ? activeSymbole : inactiveSymbole);
    display.display();

    counter++;
  }
  if (timer + 10000 < millis()){
    
    Serial.println();
    Serial.println("Connecting to wifi failed - continuing with old data");
    display.clear();
    display.setFont(ArialMT_Plain_10);
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(60, 10, "Connection to WiFi failed");
    display.drawString(60, 20, "Try again in 10mins");
    display.drawString(60, 30, "Continue with old data");
    display.display();
    delay (6000);
    readyForWeatherUpdate = false;
   }
}


